function dataOut = ...
               addGhostMixedDirichletExtrapolate(dataIn, dim, width, ghostData)
% addGhostMixedDirichletExtrapolate: mixed boundary condition ghost cells.
%
%   dataOut = addGhostMixedDirichletExtrapolate(dataIn, dim, width, ghostData)
%
% Creates ghost cells to manage the boundary conditions for the array dataIn.
%
% This BC m-file is specially designed for the TCP communication example.
%
% The boundary conditions for this 1D example are mixed:
%
% At the lower boundary (which should be zero) 
%   the BC are homogenous Dirichlet.
%
% At the upper boundary the BC are extrapolated -- either linearly or
%   quadratically.  At present, the extrapolation is done with no regard 
%   to whether the implicit surface function is bending toward or away from 
%   a (nonexistant) implicit surface beyond the computational domain boundary.
%
% It would be nice to find a theoretically justified fix,
%   which would hopefully also explain why the sign corrected extrapolation 
%   in addGhostExtrapolate works so well.
%
% Notice that the indexing is shifted by the ghost cell width in output array.
%   So in 2D with dim == 1, the first data in the original array will be at
%          dataOut(width+1,1) == dataIn(1,1)
%
% parameters:
%   dataIn	Input data array.
%   dim		Dimension in which to add ghost cells.
%   width	Number of ghost cells to add on each side (default = 1).
%   ghostData	A structure (see below).
%
%   dataOut	Output data array.
%
% ghostData is a structure containing data specific to this type of
%   ghost cell.  For this function it contains the field(s):
%
%   .order      Integer order of extrapolation, either 1 or 2.
%                 Defaults to 1 (linear).

% Copyright 2004 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Ian Mitchell, 12/11/04

%---------------------------------------------------------------------------
% Check input parameters.

if(nargin < 3)
  width = 1;
end

if((width < 0) || (width > size(dataIn, dim)))
  error('Illegal width parameter');
end

%---------------------------------------------------------------------------
% Create cell arrays with array sizes for indexing.
dims = ndims(dataIn);
sizeIn = size(dataIn);
indicesOut = cell(dims, 1);
for i = 1 : dims
  indicesOut{i} = 1:sizeIn(i);
end
indicesMid = indicesOut;
indicesIn = indicesOut;

%---------------------------------------------------------------------------
% Create appropriately sized output array.
sizeOut = sizeIn;
sizeOut(dim) = sizeOut(dim) + 2 * width;
dataOut = zeros(sizeOut);

% Fill output array with input data from the interior of the domain.
indicesOut{dim} = width + 1 : sizeOut(dim) - width;
dataOut(indicesOut{:}) = dataIn;

%---------------------------------------------------------------------------
% The lower boundary is always homogenous Dirichlet.
lowerValue = 0;
indicesOut{dim} = 1 : width;
dataOut(indicesOut{:}) = lowerValue;

%---------------------------------------------------------------------------
% The upper boundary is extrapolated.

%---------------------------------------------------------------------------
% Compute slopes (first order approximations).
indicesOut{dim} = sizeIn(dim);
indicesIn{dim} = sizeIn(dim) - 1;
slopeTop = dataIn(indicesOut{:}) - dataIn(indicesIn{:});


%---------------------------------------------------------------------------
% Final set of code depends on order of extrapolation.

if((nargin > 3) && isfield(ghostData, 'order') && (ghostData.order == 2))
  % Quadratic extrapolation.
  % No attempt to stabilize.

  % Compute second order approximation.
  indicesOut{dim} = sizeIn(dim);
  indicesMid{dim} = sizeIn(dim) - 1;
  indicesIn{dim} = sizeIn(dim) - 2;
  secondTop = dataIn(indicesOut{:}) - 2 * dataIn(indicesMid{:}) ...
              + dataIn(indicesIn{:});


else

  % Linear extrapolation.  No need for second order term.
  secondTop = 0;

  % Stabilize BC by enforcing slope that points toward zero?
  forceTowardZero = 0;
  if(forceTowardZero)
    slopeMultiplier = -1;
    indicesIn{dim} = sizeIn(dim);
    slopeTop = slopeMultiplier * abs(slopeTop) .* sign(dataIn(indicesIn{:}));
  end

end

%---------------------------------------------------------------------------
% Extrapolation to fill in ghost cells.
for i = 1 : width
  gap = width - i + 1;

  indicesOut{dim} = sizeOut(dim) - i + 1;
  indicesIn{dim} = sizeIn(dim);
  dataOut(indicesOut{:}) = (dataIn(indicesIn{:}) + gap * slopeTop ...
                            + 0.5 * gap * (gap + 1) * secondTop);
end
